# -*- coding: utf-8 -*-
#
# Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
# Author: John Hampton <pacopablo@pacopablo.com>

# Standard Library Imports
import logging
import logging.handlers
import sys

# Third Party Imports

# Local Imports


__all__ = [
    'get_logger',
    'BufferHandler',
    'NullLogger',
    'NullHandler',
]

class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def NullLogger():
    logger = logging.getLogger('dustbowl')
    logger.setLevel(logging.CRITICAL)
    logger.addHandler(NullHandler())
    return logger


class BufferHandler(logging.Handler):
    """ Simple Log handler that buffers messages into a list """
    def __init__(self, level=logging.NOTSET):
        logging.Handler.__init__(self, level)
        self.buffer = []

    def emit(self, record):
        """ Add the message to the buffer """
        try:
            self.buffer.extend(self.format(record).split('\n'))
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def get_length(self):
        return len(self.buffer)

    def get_lines(self, start=0, end=-1):
        if end < 0:
            return self.buffer[start:]
        else:
            return self.buffer[start:end]

    def clear_log(self):
        self.buffer = []


def get_logger(logname='dustbowl', level=logging.DEBUG, format=''):
    logger = logging.getLogger(logname)
    logger.setLevel(level)
    hndlr = BufferHandler()
    if format:
        fmt = logging.Formatter(format)
        hndlr.setFormatter(fmt)
    logger.addHandler(hndlr)
    logger._buffer_hndlr = hndlr
    return logger



# -*- coding: utf-8 -*-
#
# Modifcations Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
#
# Original code from Trac trunk r8199 
# http://svn.edgewall.org/reops/trac/trunk
#
# Moved from core.py
#
# Original Copyright and License:
# Copyright (C) 2003-2009 Edgewall Software
# Copyright (C) 2003-2004 Jonas Borgström <jonas@edgewall.com>
# Copyright (C) 2004-2005 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://trac.edgewall.org/log/.
#
# Author: Jonas Borgström <jonas@edgewall.com>
#         Christopher Lenz <cmlenz@gmx.de>

# Local Imports
from error import DustbowlError

__all__ = [
    'Component',
    'ExtensionPoint',
    'implements',
    'Interface',
    'IShellCommandProvider',
    'IShellConsoleObjectProvider',
    'DustbowlObj',
]


class Interface(object):
    """Marker base class for extension point interfaces."""


class ExtensionPoint(property):
    """Marker class for extension points in components."""

    def __init__(self, interface):
        """Create the extension point.

        @param interface: the `Interface` subclass that defines the protocol
            for the extension point
        """
        property.__init__(self, self.extensions)
        self.interface = interface
        self.__doc__ = 'List of components that implement `%s`' % \
                       self.interface.__name__

    def extensions(self, component):
        """Return a list of components that declare to implement the extension
        point interface.
        """
        classes = ComponentMeta._registry.get(self.interface, [])
        return filter(None, [component.compmgr[cls] for cls in classes])

    def __repr__(self):
        """Return a textual representation of the extension point."""
        return '<ExtensionPoint %s>' % self.interface.__name__


class ComponentMeta(type):
    """Meta class for components.

    Takes care of component and extension point registration.
    """
    _components = []
    _registry = {}

    def __new__(cls, name, bases, d):
        """Create the component class."""

        new_class = type.__new__(cls, name, bases, d)
        if name == 'Component':
            # Don't put the Component base class in the registry
            return new_class

        # Only override __init__ for Components not inheriting ComponentManager
        if True not in [issubclass(x, ComponentManager) for x in bases]:
            # Allow components to have a no-argument initializer so that
            # they don't need to worry about accepting the component manager
            # as argument and invoking the super-class initializer
            init = d.get('__init__')
            if not init:
                # Because we're replacing the initializer, we need to make sure
                # that any inherited initializers are also called.
                for init in [b.__init__._original for b in new_class.mro()
                             if issubclass(b, Component)
                             and '__init__' in b.__dict__]:
                    break
            def maybe_init(self, compmgr, init=init, cls=new_class):
                if cls not in compmgr.components:
                    compmgr.components[cls] = self
                    if init:
                        try:
                            init(self)
                        except:
                            del compmgr.components[cls]
                            raise
            maybe_init._original = init
            new_class.__init__ = maybe_init

        if d.get('abstract'):
            # Don't put abstract component classes in the registry
            return new_class

        ComponentMeta._components.append(new_class)
        registry = ComponentMeta._registry
        for interface in d.get('_implements', []):
            registry.setdefault(interface, []).append(new_class)
        for base in [base for base in bases if hasattr(base, '_implements')]:
            for interface in base._implements:
                registry.setdefault(interface, []).append(new_class)

        return new_class


class Component(object):
    """Base class for components.

    Every component can declare what extension points it provides, as well as
    what extension points of other components it extends.
    """
    __metaclass__ = ComponentMeta

    def __new__(cls, *args, **kwargs):
        """Return an existing instance of the component if it has already been
        activated, otherwise create a new instance.
        """
        # If this component is also the component manager, just invoke that
        if issubclass(cls, ComponentManager):
            self = super(Component, cls).__new__(cls)
            self.compmgr = self
            return self

        # The normal case where the component is not also the component manager
        compmgr = args[0]
        self = compmgr.components.get(cls)
        if self is None:
            self = super(Component, cls).__new__(cls)
            self.compmgr = compmgr
            compmgr.component_activated(self)
        return self

    @staticmethod
    def implements(*interfaces):
        """Can be used in the class definition of `Component` subclasses to
        declare the extension points that are extended.
        """
        import sys

        frame = sys._getframe(1)
        locals_ = frame.f_locals

        # Some sanity checks
        assert locals_ is not frame.f_globals and '__module__' in locals_, \
               'implements() can only be used in a class definition'

        locals_.setdefault('_implements', []).extend(interfaces)


implements = Component.implements


class ComponentManager(object):
    """The component manager keeps a pool of active components."""

    def __init__(self):
        """Initialize the component manager."""
        self.components = {}
        self.enabled = {}
        if isinstance(self, Component):
            self.components[self.__class__] = self

    def __contains__(self, cls):
        """Return whether the given class is in the list of active components"""
        return cls in self.components

    def __getitem__(self, cls):
        """Activate the component instance for the given class, or return the
        existing the instance if the component has already been activated.
        """
        if cls not in self.enabled:
            self.enabled[cls] = self.is_component_enabled(cls)
        if not self.enabled[cls]:
            return None
        component = self.components.get(cls)
        if not component:
            if cls not in ComponentMeta._components:
                raise DustbowlError('Component "%s" not registered' % cls.__name__)
            try:
                component = cls(self)
            except TypeError, e:
                raise DustbowlError('Unable to instantiate component %r (%s)' %
                                (cls, e))
        return component

    def disable_component(self, component):
        """Force a component to be disabled.

        The argument `component` can be a class or an instance.
        """
        if not isinstance(component, type):
            component = component.__class__
        self.enabled[component] = False
        self.components[component] = None

    def component_activated(self, component):
        """Can be overridden by sub-classes so that special initialization for
        components can be provided.
        """

    def is_component_enabled(self, cls):
        """Can be overridden by sub-classes to veto the activation of a
        component.

        If this method returns False, the component with the given class will
        not be available.
        """
        return True


class IShellCommandProvider(Interface):
    def match(self, cmd):
        """ Returns True if the provider handles the provided command.  False otherwise. """

    def run(self, cmd, *args, **kwargs):
        """ Invokes the given command """


class IShellConsoleObjectProvider(Interface):
    def get_console_objects(self):
        """ Return an iterable of 2-tuples of (key, value) pairs.

        The value will be injected in the console context as the given key
        """

class DustbowlObj(object):
    """Blank object used in providing namespaced console objects"""
    pass



# -*- coding: utf-8 -*-
#
# Modifcations Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
#
# Original code from Trac trunk r8199
# http://svn.edgewall.org/reops/trac/trunk
#
# Moved from core.py
#
# Original Copyright and License:
# Copyright (C) 2003-2009 Edgewall Software
# Copyright (C) 2003-2004 Jonas Borgström <jonas@edgewall.com>
# Copyright (C) 2004-2005 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://trac.edgewall.org/log/.
#
# Author: Jonas Borgström <jonas@edgewall.com>
#         Christopher Lenz <cmlenz@gmx.de>

__all__ = [
    'DustbowlError',
    'ConfigurationError',
    'ConsoleObjectError',
]

class DustbowlError(Exception):
    """Exception base class for errors in Gartersnake."""

    title = 'Gartersnake Error'

    def __init__(self, message, title=None, show_traceback=False):
        Exception.__init__(self, message)
        self.message = message
        if title:
            self.title = title
        self.show_traceback = show_traceback

    def __unicode__(self):
        return unicode(self.message)

class ConfigurationError(DustbowlError):
    """Exception raised when a value in the configuration file is not valid."""
    title = 'Configuration Error'


class ConsoleObjectError(Exception):
    """Raised when unable to assign a console object.

    This may be due to the console object already existing, or the namespace
    not being able to support the object.
    """

    def __init__(self, key, is_ns=False, cur_ns=None):
        if not is_ns or (is_ns and not cur_ns):
            self.msg = "The key >>%s<< already exists in the console." % key
        else:
            self.msg = "The namespace %s is not an object that can be " \
                       "assigned methods and other namespaces.  Can not " \
                       "add object: %s" % ('.'.join(cur_ns), key)
        self.key = key
        self.is_ns = is_ns
        self.cur_ns = cur_ns or []

    def __repr__(self):
        return "<ConsoleObjectError(key=%s, is_ns=%s, cur_ns=%s)>" \
               % (str(self.key), str(self.is_ns), str(self.cur_ns))




# -*- coding: utf-8 -*-
#
# Modifcations Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
#
# Original code from Trac trunk r8199
# http://svn.edgewall.org/reops/trac/trunk
#
# Moved from util/text.py
#
# Copyright (C) 2003-2009 Edgewall Software
# Copyright (C) 2003-2004 Jonas Borgström <jonas@edgewall.com>
# Copyright (C) 2006 Matthew Good <trac@matt-good.net>
# Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://trac.edgewall.org/log/.
#
# Author: Jonas Borgström <jonas@edgewall.com>
#         Matthew Good <trac@matt-good.net>
#         Christian Boos <cboos@neuf.fr>

# Standard Library Imports
import locale
import traceback
from StringIO import StringIO

# Third Party Imports

# Local Imports

__all__ = [
    'to_unicode',
    'format_exception',
    'get_last_traceback',
]

def to_unicode(text, charset=None):
    """Convert a `str` object to an `unicode` object.

    If `charset` is given, we simply assume that encoding for the text,
    but we'll use the "replace" mode so that the decoding will always
    succeed.
    If `charset` is ''not'' specified, we'll make some guesses, first
    trying the UTF-8 encoding, then trying the locale preferred encoding,
    in "replace" mode. This differs from the `unicode` builtin, which
    by default uses the locale preferred encoding, in 'strict' mode,
    and is therefore prompt to raise `UnicodeDecodeError`s.

    Because of the "replace" mode, the original content might be altered.
    If this is not what is wanted, one could map the original byte content
    by using an encoding which maps each byte of the input to an unicode
    character, e.g. by doing `unicode(text, 'iso-8859-1')`.
    """
    if charset:
        return unicode(text, charset, 'replace')
    else:
        try:
            return unicode(text, 'utf-8')
        except UnicodeError:
            return unicode(text, locale.getpreferredencoding(), 'replace')

def format_exception(e, traceback=""):
    message = '%s: %s' % (e.__class__.__name__, str(e))
    if traceback:
        traceback_only = get_last_traceback().split('\n')[:-2]
        message = '\n%s\n%s' % ('\n'.join(traceback_only), message)
    return message

def get_last_traceback():
    tb = StringIO()
    traceback.print_exc(file=tb)
    return tb.getvalue()




# -*- coding: utf-8 -*-
#
# Modifcations Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
#
# Original code from Trac trunk r8199
# http://svn.edgewall.org/reops/trac/trunk
#
# Original Copyright and License:
# Copyright (C) 2003-2009 Edgewall Software
# Copyright (C) 2003-2007 Jonas Borgström <jonas@edgewall.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://trac.edgewall.org/log/.
#
# Author: Jonas Borgström <jonas@edgewall.com>

# Standard Library imports
import sys
import os.path
import inspect

# Third Party imports
import pkg_resources
from pkg_resources import DistributionNotFound, VersionConflict, UnknownExtra

# Local imports
from api import Component, ComponentManager, Interface, IShellCommandProvider
from api import ExtensionPoint, IShellConsoleObjectProvider, DustbowlObj
from error import ConsoleObjectError
from config import Configuration
from log import NullLogger
from util import format_exception

__all__ = [
    'Environment',
    'IEnvObjectProvider',
]


class IEnvObjectProvider(Interface):
    def get_console_objects(self):
        """ Return an iterable of 2-tuples of (key, value) pairs.

        The value will be attached to the environment as the given key
        """


class Environment(Component, ComponentManager):
    """The environment loads plugins """

    commands = ExtensionPoint(IShellCommandProvider)
    console_objects = ExtensionPoint(IShellConsoleObjectProvider)
    env_objects = ExtensionPoint(IEnvObjectProvider)

    def __init__(self, config=None, entry_point=None, plugins=None,
                logger=None, locals=None):
        """Initialize the Dustbowl environment.

        @param config: the absolute path to a configuration file.
        @param entry_point: The entry point used to locate plugins.
        @param plugins: a list of tuples containing paths in which to look for
                        plugins and a boolean indicating whether or not
                        plugins loaded from the specified path should be
                        auto-enabled.
        @param logger: a Python logger instance.

        ``sys.path`` will be automatically added to the list of plugin
        directories.  All entries of ``sys.path`` will not be auto-enabled.
        """
        ComponentManager.__init__(self)


        self.setup_config(config)
        self.setup_log(logger)

        # Load plugins
        self.plugin_data = dict()
        if not plugins:
            plugins = list()
        self.load_modules(plugins, entry_point=entry_point)

        if locals:
            self.parent_locals = locals

        for provider in self.console_objects:
            for key, value in provider.get_console_objects():
                self.add_console_object(key, value, provider.__class__.__name__)
                continue
            continue

        for provider in self.env_objects:
            for key, value in provider.get_env_objects():
                self.add_env_object(key, value, provider.__class__.__name__)
                continue
            continue


    def component_activated(self, component):
        """Initialize additional member variables for components.

        Every component activated through the `Environment` object gets three
        member variables: `env` (the environment object), `config` (the
        environment configuration) and `log` (a logger object)."""
        component.env = self
        component.config = self.config
        component.log = self.log

    def is_component_enabled(self, cls):
        """Implemented to only allow activation of components that are not
        disabled in the configuration.

        This is called by the `ComponentManager` base class when a component is
        about to be activated. If this method returns false, the component does
        not get activated."""
        if not isinstance(cls, basestring):
            component_name = (cls.__module__ + '.' + cls.__name__).lower()
        else:
            component_name = cls.lower()

        for key, value in self.plugin_data.iteritems():
            if component_name == key or component_name.startswith(key + '.'):
                value['activated'] = True
                return value['loaded']
        return False

    def setup_config(self, configpath):
        """Load the configuration file."""
        self.config = Configuration(configpath)

    def setup_log(self, logger):
        """Initialize the logging sub-system."""
        if logger:
            self.log = logger
        else:
            self.log = NullLogger()

    def is_enabled(self, module_name):
        """ Return whether a module is enabled in the config.
        """
        for key, value in self.config.options('components'):
            k = key.lower()
            mod = module_name.lower()
            if mod == k or k.endswith('*') and mod.startswith(k[:-1]):
                return self.config.getbool('components', key)
        return False

    def load_modules(self, plugins=None, entry_point='dustbowl.modules'):
        """ Load plugins """
        def _log_error(item, e):
            ue = format_exception(e)
            if isinstance(e, DistributionNotFound):
                self.log.debug('Skipping "%s": ("%s" not found)', item, ue)
            elif isinstance(e, VersionConflict):
                self.log.error('Skipping "%s": (version conflict "%s")',
                              item, ue)
            elif isinstance(e, UnknownExtra):
                self.log.error('Skipping "%s": (unknown extra "%s")', item, ue)
            elif isinstance(e, ImportError):
                self.log.error('Skipping "%s": (can\'t import "%s")', item, ue)
            else:
                self.log.error('Skipping "%s": (error "%s")', item, ue)

        ws = pkg_resources.WorkingSet(sys.path)

        # Look for core modules
        ws.add_entry(os.path.dirname(
                        pkg_resources.resource_filename(__name__, '')))
        for entry in ws.iter_entry_points(entry_point):
            if entry.name.startswith('dustbowl.plugins'):
                entry_data = {
                    'entry' : entry,
                    'loaded' : False,
                    'activated' : False,
                    'auto_enable' : True,
                }
                self.plugin_data.setdefault(entry.name, entry_data)

        # Look for modules in sys.path
        for p in sys.path:
            ws.add_entry(p)
        for entry in ws.iter_entry_points(entry_point):
            entry_data = {
                'entry' : entry,
                'loaded' : False,
                'activated' : False,
                'auto_enable' : False,
            }
            self.plugin_data.setdefault(entry.name, entry_data)

        # Look for modules from plugin dir specified in the config or on the
        # command line
        distributions, errors = ws.find_plugins(
                                        pkg_resources.Environment(plugins))
        for dist in distributions:
            self.log.debug('Found plugin %s at %s', dist, dist.location)
            ws.add(dist)

        for entry in ws.iter_entry_points(entry_point):
            entry_data = {
                'entry' : entry,
                'loaded' : False,
                'activated' : False,
                'auto_enable' : False,
                }
            self.plugin_data.setdefault(entry.name, entry_data)

        for entry_name, data in self.plugin_data.iteritems():
            if data['auto_enable'] or self.is_enabled(entry_name):
                try:
                    self.log.debug('Loading %s from %s', data['entry'].name,
                                  data['entry'].dist)
                    # We need to make sure the distribution is on the path
                    # before we can load it.
                    data['entry'].dist.activate()
                    data['entry'].load(require=True)
                    data['loaded'] = True
                except (ImportError, DistributionNotFound, VersionConflict,
                        UnknownExtra), e:
                    # Print the last traceback to the debug buffer
                    _log_error(data['entry'], e)
            continue


        for dist, e in errors.iteritems():
            _log_error(dist, e)


    #noinspection PyBroadException
    def __call__(self, cmd, *args, **kwargs):
        """ Invoke a command """
        try:
            for command in self.commands:
                if command.match(cmd):
                    command.run(cmd, *args, **kwargs)
                    break
                continue
            else:
                print("Could not find implementation for %s" % str(cmd))
        except:
            print("Error when calling %s.  See the log for details" % str(cmd))
            self.log.error("Error calling %s" % str(cmd), exc_info=True)

    def _add_namespace_console_object(self, parts, value):
        """ Return the namesapce object with the value attached.

        Handles sub-namespaces also
        """

        key = parts[-1]
        namespaces = parts[:-1]
        global_ns = self.parent_locals.get(namespaces[0], DustbowlObj())
        ns_obj = global_ns
        cur_ns = [namespaces[0]]
        for ns in namespaces[1:]:
            cur_ns.append(ns)
            if not hasattr(ns_obj, ns):
                setattr(ns_obj, ns, DustbowlObj())
            ns_obj = getattr(ns_obj, ns)
            if not isinstance(ns_obj, DustbowlObj):
                raise ConsoleObjectError('.'.join(parts), True, cur_ns)
        if not hasattr(ns_obj, key):
            setattr(ns_obj, key, value)
        else:
            raise ConsoleObjectError('.'.join(parts))
        self.parent_locals[namespaces[0]] = global_ns


    def _add_global_console_object(self, key, value):
        """ Adds a console object without considering namespaces """
        if key in self.parent_locals:
            raise ConsoleObjectError(key)
        self.parent_locals[key] = value


    def add_console_object(self, key, value, provider=None):
        """ Add the value to the console context with the given key """
        if provider is None:
            caller_frame = inspect.stack()[1][0]
            caller_class = ''
            if 'self' in caller_frame.f_locals:
                caller_class = caller_frame.f_locals['self'].__class__.__name__
            provider = "%s%s" % (caller_class and (caller_class + '.') or '',
                                 caller_frame.f_code.co_name)

        parts = key.split('.')
        is_ns = len(parts) > 1
        try:
            if not is_ns:
                self._add_global_console_object(key, value)
            else:
                self._add_namespace_console_object(parts, value)
            self.log.debug("%s added '%s' to the console" %
                           (provider, '.'.join(parts) +
                           (callable(value) and '()' or '')))
        except ConsoleObjectError, e:
            self.log.error("%s  %s must provide a different key"
                           % (e.msg, provider))
            return


    def add_env_object(self, key, value, provider=None):
        """ Add the value to the environment as the given attribute """
        if provider is None:
            caller_frame = inspect.stack()[1][0]
            caller_class = ''
            if 'self' in caller_frame.f_locals:
                caller_class = caller_frame.f_locals['self'].__class__.__name__ 
            provider = "%s%s" % (caller_class and (caller_class + '.') or '',
                                 caller_frame.f_code.co_name)

        attr = getattr(self, key, None)
        if attr:
            self.log.error("The attribute/method >>%s<< already exists in the "
                           "environment.  %s must provide a different "
                           "key" % (str(key), provider))
        else:
            setattr(self, key, value)
            self.log.debug("%s added '%s' to the environemnt" %
                           (provider,
                            str(key) + (callable(value) and '()' or '')))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
# Author: John Hampton <pacopablo@pacopablo.com>

# Standard Library Imports

# Third Party Imports

# Local Imports
from dustbowl.api import IShellConsoleObjectProvider, Component, implements

__all__ = [
    'ConfigManager',
]

class ConfigManager(Component):
    """ Provides access ot the config object """

    implements(IShellConsoleObjectProvider)

    def get_console_objects(self):
        """ Add access to the config object.

        TODO: This should probably be replaced by a plugin that provides a
        nicer interface to the config data
        """
        yield ('shellconfig', self.config)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
# Author: John Hampton <pacopablo@pacopablo.com>

# Standard Library Imports
import logging

# Third Party Imports

# Local Imports
from dustbowl.api import IShellCommandProvider, Component, implements

__all__ = [
    'LogCmdProvider',
]

#noinspection PyInitNewSignature
class LogCmdProvider(Component):
    """ Shows the messages that have been logged.

    The log command requires one of the following arguements:
     * show
     * save
     * clear

    show:  Will show all unread log messages.  If passed the 'all' parameter, all
           log message with be shown regardless of read status.
    save:  Must be passed the name of a file.  The log will be saved to the
           specified file.  The file will be overwritten if it already exists.
    clear: Clears the log.

    Examples:
     1) >>> .log.show
     1) >>> .log.show.all
     1) >>> .log.save 'logfile.log'
     1) >>> .log.clear
    """

    implements(IShellCommandProvider)

    def __init__(self):
        self.level = self.log.getEffectiveLevel()

    def match(self, cmd):
        c = cmd.lower()
        return c == 'log' or c.startswith('log.')

    def run(self, cmd, *args, **kwargs):
        cmds = cmd.split('.')
        num_cmds = len(cmds)
        if num_cmds == 1:
            self.log.log(self.level, *args, **kwargs)
        elif num_cmds > 1:
            f = getattr(self, '_log_%s' % cmds[1], None)
            if f:
                f(cmd, *args, **kwargs)

    def _log_setlevel(self, *args, **kwargs):
        if len(*args) < 1:
            print("A logging level of DEBUG, INFO, WARNING, ERROR, or CRITICAL must be specified.")
            print("The current log level is: %s" % logging._levelNames[self.level])
        else:
            level = args[0]
            if isinstance(level, basestring) and \
               level.upper() in logging._levelNames:
                self.level = logging._levelNames[level]
            elif isinstance(level, int) and level in logging._levelNames:
                self.level = level
            else:
                print("Unknown logging level %s.  Log level remains: %s" %
                        (level, logging._levelNames[self.level]))
        pass

    def _log_show(self, cmd, *args, **kwargs):
        """ Show log entries in the console.

        When invoked as .log.show, all log entries that have not yet been
        viewed are shown.  If invoked as .log.show.all, the entire log is
        shown, regardless of whether it has been seen previously.
        """
        startidx = getattr(self, 'startidx', 0)
        length = self.log._buffer_hndlr.get_length()

        cmds = cmd.split('.')
        showall = cmds[2:3]
        if not showall:
            text = '\n  '.join(self.log._buffer_hndlr.get_lines(startidx))
        elif showall[0].lower() == 'all':
            text = '\n  '.join(self.log._buffer_hndlr.get_lines())
        else:
            text = '\n'
        print('  ' + text)
        self.startidx = length
        return False

    def _log_save(self, *args, **kwargs):
        print('Not Implemented yet')
        return False

    def _log_clear(self, *args, **kwargs):
        self.log._buffer_hndlr.clear_log()
        self.startidx = 0
        print('Log Cleared')
        return False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 John Hampton <pacopablo@pacopablo.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
# Author: John Hampton <pacopablo@pacopablo.com>

__author__ = 'John Hampton'

# Std Library Imports

# Local Imports

# Third Party Imports
from dustbowl.api import IShellCommandProvider, Component, implements
from colorama import Fore, Style


__all__ = [
    'ModuleCmdProvider',
    ]

#noinspection PyInitNewSignature
class ModuleCmdProvider(Component):
    """ Enables management of plugin modules.

    The module command requires one of the following arguments:
     * list
     * enable
     * disable

    list:  Will show all unread log messages.  If passed the 'all' parameter,
    all
           log message with be shown regardless of read status.
    enable:  Must be passed the name of a file.  The log will be saved to the
           specified file.  The file will be overwritten if it already exists.
    disable: Clears the log.

    Examples:
     1) >>> .module.list
     1) >>> .module.list 'all'
     1) >>> .module.list 'enabled'
     1) >>> .module.list 'disabled'
     1) >>> .module enable 'module_spec'
     1) >>> .module.disable 'nodule_spec'
    """

    implements(IShellCommandProvider)

    def match(self, cmd):
        c = cmd.lower()
        return c == 'module' or c.startswith('module.')

    def run(self, cmd, *args, **kwargs):
        cmds = cmd.split('.')
        num_cmds = len(cmds)
        if num_cmds == 1:
            cmd = 'list'
        else:
            cmd = cmds[1]
        self.log.debug('ModuleCmdProvider: num_cmds = %d' % num_cmds)
        f = getattr(self, '_module_%s' % cmd, None)
        if f:
            f(cmd, *args, **kwargs)

    def _module_list(self, cmd, *args, **kwargs):
        """ List all modules found and their activation status
        """

        print ("_module_list args: %s" % str(args))
        if len(args) > 0:
            list_option = str(args[0]).lower()
        else:
            list_option = 'all'
        print (Style.DIM + "Active   Module" + Style.NORMAL)
        sorted_modules = self.env.plugin_data.keys()
        sorted_modules.sort()
        for name in sorted_modules:
            if self.env.plugin_data[name]['activated']:
                if list_option in ['all', 'enabled']:
                    print(Fore.GREEN + Style.BRIGHT + "  *      " +
                          Style.RESET_ALL + "%s" % name)
            else:
                if list_option in ['all', 'disabled']:
                    print(Style.DIM + "         %s" % name + Style.NORMAL)
            continue







#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
# Author: John Hampton <pacopablo@pacopablo.com>

from setuptools import setup

setup(
    name='Dustbowl',
    version='1.0.1',
    packages=['dustbowl', 'dustbowl.plugins'],
    author='John Hampton',
    author_email='pacopablo@pacopablo.com',
    description='Dustbowl',
    scripts=['scripts/dustbowl'],
    url='http://pacopablo.github.com/dustbowl',
    install_requires = ['SQLAlchemy', 'colorama'],
    license='MIT',
    zip_safe = False,
    entry_points = """
        [dustbowl.modules]
        dustbowl.plugins.logcmd = dustbowl.plugins.logcmd
        dustbowl.plugins.datasources = dustbowl.plugins.datasources
        dustbowl.plugins.config = dustbowl.plugins.config
        dustbowl.plugins.modules = dustbowl.plugins.modules
    """,
    long_description = """
    Dustbowl
    ==========

    Dustbowl is an InteractiveConsole_ that allows for loading plugins.  These
    plugins are python packages that ``[dustbowl.modules]`` entry points.  One
    can create plugins for commonly repeated tasks.  The benefit to using an
    InteractiveConsole_ is that once a method from a plugin has been run, one
    can further inspect and manipulate its results and data.  Thus we have all
    the power of the Python REPL with assisting functions and commands.

    .. Links
    .. _InteractiveConsole: http://docs.python.org/library/code.html#code.InteractiveConsole
    """

)




# -*- coding: utf-8 -*-
#
# Copyright (C) 2009, John Hampton <pacopablo@pacopablo.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
# Author: John Hampton <pacopablo@pacopablo.com>

# Standard library imports
import sys
from code import InteractiveConsole, compile_command

# Third Party imports
import colorama
from colorama import Fore, Style

# Local imports
import dustbowl.env

VERSION = '1.0.1'

CMD_TOKEN = "."

class DustbowlConsole(InteractiveConsole):

    def __init__(self, locals=None, filename="<console>", args=None, logger=None):
        InteractiveConsole.__init__(self, locals=locals, filename=filename)
        self.args = args
        self.locals = locals
        self.env = dustbowl.env.Environment(args.config,
                                          'dustbowl.modules',
                                          args.plugins,
                                          logger,
                                          self.locals)
        self.locals['__env__'] = self.env

    def interact(self, banner=None):
        global CMD_TOKEN

        # So, these two try statements don't seem very Pythonic.  I'm game
        # for a nicer way to determine whether or not ``ps1`` and ``ps2`` exist
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        cprt = 'Type "help", "copyright", "credits" or "license" for more information.'
        colorama.init()
        if banner is None:
            self.write(Style.BRIGHT + Fore.BLUE + "Dustbowl" + Fore.RESET +
                       Style.NORMAL + ": Python %s on %s\n%s\n" %
                       (sys.version, sys.platform, cprt))
        else:
            self.write("%s\n" % str(banner))
        for i in self.env.plugin_data.itervalues():
            if i['loaded']:
                self.write(Style.DIM + "Loaded %s\n" % i['entry'].name +
                           Style.NORMAL)
        if 'readline' in sys.modules:
            self.write("Dustbowl Tabbed Completion Enabled\n")

        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = self.raw_input(prompt)
                    # Can be None if sys.stdin was redefined
                    encoding = getattr(sys.stdin, "encoding", None)
                    if encoding and not isinstance(line, unicode):
                        line = line.decode(encoding)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    if line.strip().startswith(CMD_TOKEN):
                        line = self.process_command(line)
                    more = self.push(line)
            except KeyboardInterrupt:
                self.write("\nKeyboardInterrupt\n")
                self.resetbuffer()
                more = 0

    def process_command(self, line):
        global CMD_TOKEN
        pos = line.find(CMD_TOKEN)
        if pos >= 0:
            parts = line[pos:].split(' ')
            cmd = parts[0]
            args = ' '.join(parts[1:])
            return '%s__env__("%s"%s)' % (line[:pos], cmd.lstrip(CMD_TOKEN), args and ', ' + args or '') 
        else:
            return line



# -*- coding: utf-8 -*-
#
# Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
# Author: John Hampton <pacopablo@pacopablo.com>

def enable_tabbed_completion(historyPath='.dustbowl.hist', context=None):
    import atexit
    import os
    import readline
    import rlcompleter

    readline.parse_and_bind('tab: complete')
    readline.parse_and_bind ("bind ^I rl_complete")
    if 'libedit' in readline.__doc__:
        readline.parse_and_bind('bind "\e[3~" ed-delete-next-char')

    context = context or globals()
    readline.set_completer(rlcompleter.Completer(context).complete)

    def save_history(historyPath=historyPath):
        import readline
        readline.write_history_file(historyPath)

    if os.path.exists(historyPath):
        readline.read_history_file(historyPath)

    atexit.register(save_history)
    del os, atexit, readline, rlcompleter, save_history, historyPath



# -*- coding: utf-8 -*-
#
# Copyright (C) 2009 John Hampton <pacopablo@pacopablo.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#
# Author: John Hampton <pacopablo@pacopablo.com>

# Standard Library Imports
import re

# Third Party Imports
from sqlalchemy import create_engine, orm
from sqlalchemy.orm import sessionmaker, scoped_session

# Local Imports
from dustbowl.api import Component, implements, IShellConsoleObjectProvider
from dustbowl.api import ExtensionPoint, Interface
from dustbowl.env import IEnvObjectProvider

__all__ = [
    'IDataSourceProvider',
    'DataSourceManager',
    'ConfigDataSource',
]

KNOWN_OPTIONS = ['encoding', 'convert_unicode']
OPTION_KEYS_RE = r'sqlalchemy\.(?P<source>[\w\d]+?)\.(?P<option>[\w\d]+)'
URL_RE = r'sqlalchemy\.(?P<source>[\w\d]+?)\.url'
url = re.compile(URL_RE)
sqloptions = re.compile(OPTION_KEYS_RE)


class IDataSourceProvider(Interface):
    def get_data_source(self):
        """ Yield (key, source) tuples. """


class DataSourceManager(Component):
    """ Provide access to datasources """

    data_source_providers = ExtensionPoint(IDataSourceProvider)
    implements(IShellConsoleObjectProvider, IEnvObjectProvider)

    def get_datasource(self, datasource, part='scoped_session', **kwargs):
        try:
            sources = self.datasources
        except AttributeError:
            self.datasources = {}
            for provider in self.data_source_providers:
                for k, engine in provider.get_data_source():
                    sm = sessionmaker(bind=engine, **kwargs)
                    ss = scoped_session(sm)
                    self.datasources[k] = {'engine': engine,
                                           'sessionmaker': sm,
                                           'scoped_session': ss,}
                    continue
                continue
            sources = self.datasources
        try:
            ret = sources[datasource][part]
        except KeyError:
            self.log.warning('The datasource >> %s << was not found' %
                             str(datasource))
            ret = None
        return ret

    def get_console_objects(self):
        yield ('get_datasource', self.get_datasource)

    def get_env_objects(self):
        yield ('get_datasource', self.get_datasource)



class ConfigDataSource(Component):
    """ Load SQLAlchemy Data Sources from the configuration file """

    implements(IDataSourceProvider)

    def get_data_source(self):
        global url

        sections = self.config.sections()
        source_info = {}
        for s in sections:
            for k, v in self.config.options(s):
                g = url.match(k)
                if g:
                    key = g.group('source')
                    args_dict = self._parse_engine_args(key, s)
                    source = create_engine(v, **args_dict)
                    source_info[key] = source
                continue
            continue

        for k, s in source_info.items():
            yield (k, s)

    def _parse_engine_args(self, source, section):
        """ Create a dictionary of keyword arguments to be passed to the
        create_engine call.

        The options are in the format: sqlalchemy.<key>.<param> = <value>

        Currently, the availalbe options are:
         * encoding
         * convert_unicode
        """
        global sqloptions, KNOWN_OPTIONS

        engine_args = {}
        for k, v in self.config.options(section):
            g = sqloptions.match(k)
            if g and (source == g.group('source')):
                option = g.group('option')
                if option in KNOWN_OPTIONS:
                    engine_args[option] = getattr(self, '_get_' + option.upper(),
                                              lambda x, y: None)(section, k)
            continue
        return engine_args

    def _get_ENCODING(self, section, option):
        return self.config.get(section, option, None)

    def _get_CONVERT_UNICODE(self, section, option):
        return self.config.getbool(section, option, False)

